@@ -36,16 +36,40 @@ use base qw(Exporter);
 sub parse {
     my ($tr, $data)   = @_;
     my $args          = $tr->parser_args;
-    my $dbicschema    = $args->{'DBIx::Class::Schema'} ||  $args->{"DBIx::Schema"} ||$data;
-    $dbicschema     ||= $args->{'package'};
-    my $limit_sources = $args->{'sources'};
+
+    my $dbicschema = $data || $args->{dbic_schema};
+
+    for (qw(DBIx::Class::Schema DBIx::Schema package)) {
+      if (defined (my $s = delete $args->{$_} )) {
+        carp_unique("Supplying a schema via  ... parser_args => { '$_' => \$schema } is deprecated. Please use parser_args => { dbic_schema => \$schema } instead");
+
+        # move it from the deprecated to the proper $args slot
+        unless ($dbicschema) {
+          $args->{dbic_schema} = $dbicschema = $s;
+        }
+      }
+    }
 
     DBIx::Class::Exception->throw('No DBIx::Class::Schema') unless ($dbicschema);
+
     if (!ref $dbicschema) {
       eval "require $dbicschema"
         or DBIx::Class::Exception->throw("Can't load $dbicschema: $@");
     }
 
+    if (
+      ref $args->{dbic_schema}
+        and
+      $args->{dbic_schema}->storage
+    ) {
+      # we have a storage-holding $schema instance in $args
+      # we need to dissociate it from that $storage
+      # otherwise SQLT insanity may ensue due to how some
+      # serializing producers treat $args (crazy crazy shit)
+      local $args->{dbic_schema}{storage};
+      $args->{dbic_schema} = $args->{dbic_schema}->clone;
+    }
+
     my $schema      = $tr->schema;
     my $table_no    = 0;
 
@@ -53,7 +77,7 @@ sub parse {
       unless ($schema->name);
 
     my @monikers = sort $dbicschema->sources;
-    if ($limit_sources) {
+    if (my $limit_sources = $args->{'sources'}) {
         my $ref = ref $limit_sources || '';
         $dbicschema->throw_exception ("'sources' parameter must be an array or hash ref")
           unless( $ref eq 'ARRAY' || ref eq 'HASH' );
@@ -144,8 +168,25 @@ sub parse {
 
             my $rel_info = $source->relationship_info($rel);
 
-            # Ignore any rel cond that isn't a straight hash
-            next unless ref $rel_info->{cond} eq 'HASH';
+            # Ignore any rel cond that isn't a straight hash or DQ expr
+
+            my $rel_cond = do {
+              if (ref($rel_info->{cond}) eq 'HASH') {
+                # strip foreign. and self.
+                +{ map {/^\w+\.(\w+)$/} %{$rel_info->{cond}} };
+              } elsif (
+                blessed($rel_info->{cond})
+                and $rel_info->{cond}->isa('Data::Query::ExprBuilder')
+              ) {
+                $source->_join_condition_to_hashref($rel_info->{cond}{expr});
+              } else {
+                undef;
+              }
+            };
+
+            # non-equality join DQ expr will also have produced undef
+
+            next unless $rel_cond;
 
             my $relsource = try { $source->related_source($rel) };
             unless ($relsource) {
@@ -167,12 +208,10 @@ sub parse {
 
             # Force the order of @cond to match the order of ->add_columns
             my $idx;
-            my %other_columns_idx = map {'foreign.'.$_ => ++$idx } $relsource->columns;
-            my @cond = sort { $other_columns_idx{$a} cmp $other_columns_idx{$b} } keys(%{$rel_info->{cond}});
+            my %other_columns_idx = map { $_ => ++$idx } $relsource->columns;
+            my @refkeys = sort { $other_columns_idx{$a} cmp $other_columns_idx{$b} } keys(%{$rel_cond});
 
-            # Get the key information, mapping off the foreign/self markers
-            my @refkeys = map {/^\w+\.(\w+)$/} @cond;
-            my @keys = map {$rel_info->{cond}->{$_} =~ /^\w+\.(\w+)$/} @cond;
+            my @keys = @{$rel_cond}{@refkeys};
 
             # determine if this relationship is a self.fk => foreign.pk (i.e. belongs_to)
             my $fk_constraint;
@@ -213,7 +252,7 @@ sub parse {
                 # Constraints are added only if applicable
                 next unless $fk_constraint;
 
-                # Make sure we dont create the same foreign key constraint twice
+                # Make sure we don't create the same foreign key constraint twice
                 my $key_test = join("\x00", sort @keys);
                 next if $created_FK_rels{$rel_table}->{$key_test};
 
@@ -422,7 +461,7 @@ from a DBIx::Class::Schema instance
  my $trans  = SQL::Translator->new (
       parser      => 'SQL::Translator::Parser::DBIx::Class',
       parser_args => {
-          package => $schema,
+          dbic_schema => $schema,
           add_fk_index => 0,
           sources => [qw/
             Artist
@@ -452,6 +491,27 @@ L<DBIx::Class::Schema/create_ddl_dir>.
 
 =head1 PARSER OPTIONS
 
+=head2 dbic_schema
+
+The DBIx::Class schema (either an instance or a class name) to be parsed.
+This argument is in fact optional - instead one can supply it later at
+translation time as an argument to L<SQL::Translator/translate>. In
+other words both of the following invocations are valid and will produce
+conceptually identical output:
+
+  my $yaml = SQL::Translator->new(
+    parser => 'SQL::Translator::Parser::DBIx::Class',
+    parser_args => {
+      dbic_schema => $schema,
+    },
+    producer => 'SQL::Translator::Producer::YAML',
+  )->translate;
+
+  my $yaml = SQL::Translator->new(
+    parser => 'SQL::Translator::Parser::DBIx::Class',
+    producer => 'SQL::Translator::Producer::YAML',
+  )->translate(data => $schema);
+
 =head2 add_fk_index
 
 Create an index for each foreign key.
